/**
 *   Copyright (C) 2012-2013 IFSTTAR (http://www.ifsttar.fr)
 *   Copyright (C) 2012-2013 Oslandia <infos@oslandia.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#ifndef TEMPUS_SPEED_PROFILE_HH
#define TEMPUS_SPEED_PROFILE_HH

#include "transport_modes.hh"

namespace Tempus {

struct SpeedTimePeriod
{
    // length of the period (minutes)
    float length;
    // average speed (km/h)
    float speed;
};

class RoadEdgeSpeedProfiles
{
public:
    RoadEdgeSpeedProfiles() {}

    ///
    /// Validity period of these profiles
    DECLARE_RW_PROPERTY( validity_period, db_id_t );

    // speed rule -> time of the day (in minutes) -> SpeedTimePeriod
    typedef std::map< TransportModeSpeedRule, std::map< float, SpeedTimePeriod > > SpeedProfile;
    // road edge -> speed profile
    typedef std::map< db_id_t, SpeedProfile > RoadEdgeSpeedProfileMap;

    typedef std::map<float, SpeedTimePeriod>::const_iterator PeriodIterator;

    ///
    /// Compute time spent on the section starting at a given time and for a given length
    /// In the normal (forward) direction
    ///
    /// Return a value <0 if there is no speed profile for this section / time
    float time_spent_on_section_forward( db_id_t section_id, TransportModeSpeedRule speed_rule, float start_time, float length ) const;

    ///
    /// Compute time spent on the section arriving at a given time and for a given length
    /// In the normal backward direction
    ///
    /// Return a value <0 if there is no speed profile for this section / time
    float time_spent_on_section_backward( db_id_t section_id, TransportModeSpeedRule speed_rule, float arriving_time, float length ) const;

    void add_period( db_id_t section_id, TransportModeSpeedRule speed_rule, float begin_time, float length, float speed );
private:
    RoadEdgeSpeedProfileMap speed_profile_;

    ///
    /// Return all the periods including and after a given time for a section and a speed rule
    std::pair<PeriodIterator, PeriodIterator> periods_after( db_id_t section_id, TransportModeSpeedRule speed_rule, float begin_time, bool& found ) const;
    std::pair<PeriodIterator, PeriodIterator> periods_before( db_id_t section_id, TransportModeSpeedRule speed_rule, float arriving_time, bool& found ) const;

    friend void serialize( std::ostream&, const RoadEdgeSpeedProfiles&, binary_serialization_t );
    friend void unserialize( std::istream&, RoadEdgeSpeedProfiles&, binary_serialization_t );
};

} // namespace Tempus

#endif
